"""
Shared constants for Biblicus.
"""

SCHEMA_VERSION = 2
COLLECTION_SCHEMA_VERSION = 1
DATASET_SCHEMA_VERSION = 1
EXTRACTION_DATASET_SCHEMA_VERSION = 1
ANALYSIS_SCHEMA_VERSION = 1
GRAPH_SCHEMA_VERSION = 1
CORPUS_DIR_NAME = "metadata"
LEGACY_CORPUS_DIR_NAME = ".biblicus"
DEFAULT_RAW_DIR = "."
SIDECAR_SUFFIX = ".biblicus.yml"
SNAPSHOTS_DIR_NAME = "snapshots"
EXTRACTION_SNAPSHOTS_DIR_NAME = "extraction"
ANALYSIS_RUNS_DIR_NAME = "analysis"
GRAPH_SNAPSHOTS_DIR_NAME = "graph"
HOOK_LOGS_DIR_NAME = "hook_logs"
EXTRACTED_DIR_NAME = "extracted"
GRAPH_DIR_NAME = "graph"
RETRIEVAL_DIR_NAME = "retrieval"
ANALYSIS_DIR_NAME = "analysis"
